<?php

namespace SuperRhino\Utils\Redis;

use SuperRhino\Utils\Snowflake\SnowflakeGenerator;

/**
 * redis锁
 */
class Lock {
    public static $lockPrefix = 'lock_';
    public static $poolName = '';

    /**
     * 加锁
     * @param string $lockKey 锁名
     * @param int $ttl        milliseconds
     * @return string|bool      锁id
     */
    public static function lock(string $lockKey, int $ttl) {
        $redis = Redis::getRedis(static::$poolName);
        $lockId = SnowflakeGenerator::generateStr();

        $result = $redis->set(static::$lockPrefix . $lockKey, $lockId, ['NX', 'PX' => $ttl]);
        if ($result) {
            return $lockId;
        }
        return false;
    }

    /**
     * 解锁
     * @param string $lockKey 锁名
     * @param string $lockId  锁id
     * @return bool
     */
    public static function unlock(string $lockKey, string $lockId) {
        $redis = Redis::getRedis(static::$poolName);
        $luaScript = <<<LUA
local value = redis.call('get', KEYS[1])
if (value == ARGV[1]) then
	return redis.call('del', KEYS[1]);
end
return 0;
LUA;
        $result = $redis->eval($luaScript, [static::$lockPrefix . $lockKey, $lockId], 1);
        return (bool)$result;
    }

    /**
     * 强制加锁 (删除旧锁，重新加锁)
     * @param string $lockKey
     * @param int $ttl
     * @return false|string
     */
    public static function forceLock(string $lockKey, int $ttl) {
        $redis = Redis::getRedis(static::$poolName);
        $lockId = SnowflakeGenerator::generateStr();

        $result = $redis->set(static::$lockPrefix . $lockKey, $lockId, ['PX' => $ttl]);
        if ($result) {
            return $lockId;
        }
        return false;
    }

    /**
     * 删除锁
     * @param string $lockKey
     * @return false|int
     */
    public static function delLock(string $lockKey) {
        $redis = Redis::getRedis(static::$poolName);
        return $redis->del(static::$lockPrefix . $lockKey);
    }

    /**
     * 重入锁(继锁)
     * @param string $lock
     * @param string $lockId
     * @param int $ttl
     * @return bool
     */
    public static function relock(string $lock, string $lockId, int $ttl) {
        $redis = Redis::getRedis(static::$poolName);
        $luaScript = <<<LUA
local value = redis.call('get', KEYS[1])
if (value == ARGV[1]) then
    return redis.call('SET', KEYS[1], ARGV[1], 'PX', ARGV[2])
end
return 0;
LUA;
        $result = $redis->eval($luaScript, [static::$lockPrefix . $lock, $lockId, $ttl], 1);
        return (bool)$result;
    }

    /**
     * 等待锁
     * @param string $lockKey
     * @param int $ttl
     * @param int $retry    重试次数
     * @param int $interval 重试间隔毫秒
     * @return bool|string
     */
    public static function lockWait(string $lockKey, int $ttl, int $retry = 5, int $interval = 20) {
        $times = 0;
        do {
            if ($lockId = Lock::lock($lockKey, $ttl)) {
                return $lockId;
            }
            $times++;
            usleep($interval * 1000);
        } while ($times < $retry);
        return false;
    }
}